// lex.cpp
// VERSION 2
// Glenn G. Chappell
// 4 Feb 2013
//
// For CS 331 Spring 2013
// Source for class Lex
// State-Machine-Based Lexical Analyzer

#include "lex.h"  // For class Lex
#include <cctype>
using std::isspace;
using std::isalpha;
using std::isdigit;
using std::isalnum;
#include <string>
using std::string;


// advance
// Moves to the next lexeme -- or ("", NONE) if there are no
//  more -- and sets currlexeme_ equal to it.
void Lex::advance()
{
    if (pos_ == input_.size())
    {
        currlexeme_ = Lexeme("", NONE);
        return;
    }

    // pos_ is now the subscript of the start of a Lexeme
    // Our task:
    // - Build that Lexeme string
    // - Figure out its token
    // - Set currlexeme_ to the right value
    // - Call skipSpace

    string lexstring = "";
    Token lextoken;

    enum State { START, DONE,
                 LETTER, DIGIT, DOT, PLUS, MINUS, EQUAL, DIGDOT };
    // Meaning of State values (same as if following has been read):
    // START:  nothing
    // DONE:   (lexeme complete; lexstring & lextoken have final values)
    // LETTER: a letter
    // DIGIT:  a digit
    // DOT:    '.'
    // PLUS:   '+'
    // MINUS:  '-'
    // EQUAL:  '='
    // DIGDOT: DIGIT, followed by '.'

    State currstate = START;

    while (currstate != DONE)
    {
        int c;
        if (pos_ == input_.size())
            c = -1;
        else
            c = input_[pos_];
        ++pos_;
        switch (currstate)
        {
        case START:
            if (isalpha(c) || c == '_')
            {
                currstate = LETTER;
            }
            else if (isdigit(c))
            {
                currstate = DIGIT;
            }
            else if (c == '.')
            {
                currstate = DOT;
            }
            else if (c == '+')
            {
                currstate = PLUS;
            }
            else if (c == '-')
            {
                currstate = MINUS;
            }
            else if (c == '=')
            {
                currstate = EQUAL;
            }
            else if (c >= ' ' && c <= '~')
            {
                currstate = DONE;
                lextoken = PUNCT;
            }
            else
            {
                currstate = DONE;
                lextoken = ILLEGAL;
            }
            lexstring.push_back(c);
            break;
        case LETTER:
            if (isalnum(c) || c == '_')
            {
                currstate = LETTER;
                lexstring.push_back(c);
            }
            else
            {
                --pos_;
                currstate = DONE;
                if (lexstring == "func" || lexstring == "return")
                    lextoken = KEY;
                else
                    lextoken = ID;
            }
            break;
        case DIGIT:
            if (isdigit(c))
            {
                currstate = DIGIT;
                lexstring.push_back(c);
            }
            else if (c == '.')
            {
                currstate = DIGDOT;
                lexstring.push_back(c);
            }
            else
            {
                --pos_;
                currstate = DONE;
                lextoken = NUM;
            }
            break;
        case DOT:
            if (isdigit(c))
            {
                currstate = DIGDOT;
                lexstring.push_back(c);
            }
            else
            {
                --pos_;
                currstate = DONE;
                lextoken = OP;
            }
            break;
        case PLUS:
            if (c == '+' || c == '=')
            {
                currstate = DONE;
                lextoken = OP;
                lexstring.push_back(c);
            }
            else if (isdigit(c))
            {
                currstate = DIGIT;
                lexstring.push_back(c);
            }
            else if (c == '.')
            {
                if (pos_ < input_.size() && isdigit(input_[pos_]))
                {
                    currstate = DIGDOT;
                    lexstring.push_back(c);
                }
                else
                {
                    --pos_;
                    currstate = DONE;
                    lextoken = OP;
                }
            }
            else
            {
                --pos_;
                currstate = DONE;
                lextoken = OP;
            }
            break;
        case MINUS:
            if (c == '-' || c == '=')
            {
                currstate = DONE;
                lextoken = OP;
                lexstring.push_back(c);
            }
            else if (isdigit(c))
            {
                currstate = DIGIT;
                lexstring.push_back(c);
            }
            else if (c == '.')
            {
                if (pos_ < input_.size() && isdigit(input_[pos_]))
                {
                    currstate = DIGDOT;
                    lexstring.push_back(c);
                }
                else
                {
                    --pos_;
                    currstate = DONE;
                    lextoken = OP;
                }
            }
            else
            {
                --pos_;
                currstate = DONE;
                lextoken = OP;
            }
            break;
        case EQUAL:
            --pos_;
            currstate = DONE;
            lextoken = OP;
            break;
        case DIGDOT:
            if (isdigit(c))
            {
                currstate = DIGDOT;
                lexstring.push_back(c);
            }
            else
            {
                --pos_;
                currstate = DONE;
                lextoken = NUM;
            }
            break;
        }
    }

    currlexeme_ = Lexeme(lexstring, lextoken);
    skipSpace();
}


// skipSpace
// Skip whitespace & comments
// Sets pos_ to the index of the start of the next lexeme after its
//  initial location, or input_.size() if there are no further
//  lexemes.
void Lex::skipSpace()
{
    while (true)
    {
        // Skip whitespace
        while (pos_ < input_.size() && isspace(input_[pos_]))
            ++pos_;
        // If not at '#', done
        if (pos_ == input_.size() || input_[pos_] != '#')
            return;
        // Found a comment; skip it
        ++pos_;
        while (pos_ != input_.size() && input_[pos_] != '\n')
            ++pos_;
    }
}

